import colander
import mock

from kinto.core.testing import unittest
from kinto.core.resource import schema


class DepracatedSchemasTest(unittest.TestCase):

    def test_resource_timestamp_is_depracated(self):
        with mock.patch('kinto.core.resource.schema.warnings') as mocked:
            schema.TimeStamp()
            message = ("`kinto.core.resource.schema.TimeStamp` is deprecated, "
                       "use `kinto.core.schema.TimeStamp` instead.")
            mocked.warn.assert_called_with(message, DeprecationWarning)

    def test_resource_URL_is_depracated(self):
        with mock.patch('kinto.core.resource.schema.warnings') as mocked:
            schema.URL()
            message = ("`kinto.core.resource.schema.URL` is deprecated, "
                       "use `kinto.core.schema.URL` instead.")
            mocked.warn.assert_called_with(message, DeprecationWarning)


class ResourceSchemaTest(unittest.TestCase):

    def test_preserves_unknown_fields_when_specified(self):
        class PreserveSchema(schema.ResourceSchema):
            class Options:
                preserve_unknown = True

        schema_instance = PreserveSchema()
        deserialized = schema_instance.deserialize({'foo': 'bar'})
        self.assertIn('foo', deserialized)
        self.assertEquals(deserialized['foo'], 'bar')

    def test_ignore_unknwon_fields_when_specified(self):
        class PreserveSchema(schema.ResourceSchema):
            class Options:
                preserve_unknown = False

        schema_instance = PreserveSchema()
        deserialized = schema_instance.deserialize({'foo': 'bar'})
        self.assertNotIn('foo', deserialized)

    def test_accepts_unknown_fields_by_default(self):
        schema_instance = schema.ResourceSchema()
        deserialized = schema_instance.deserialize({'foo': 'bar'})
        self.assertIn('foo', deserialized)

    def test_options_parameters_use_default_value_when_subclassed(self):
        class PreserveSchema(schema.ResourceSchema):
            class Options:
                pass

        schema_instance = PreserveSchema()
        deserialized = schema_instance.deserialize({'foo': 'bar'})
        self.assertIn('foo', deserialized)


class PermissionsSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.PermissionsSchema()

    def test_works_with_any_permission_name_by_default(self):
        perms = {'can_cook': ['mat']}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_works_with_empty_mapping(self):
        perms = {}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_works_with_empty_list_of_principals(self):
        perms = {'can_cook': []}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_raises_invalid_if_not_a_mapping(self):
        perms = ['gab']
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          perms)

    def test_raises_invalid_if_permission_is_unknown(self):
        self.schema.known_perms = ('can_sleep',)
        perms = {'can_work': ['mat']}
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          perms)

    def test_raises_invalid_if_not_list(self):
        perms = {'can_cook': 3.14}
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          perms)

    def test_raises_invalid_if_not_list_of_strings(self):
        perms = {'can_cook': ['pi', 3.14]}
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          perms)


class HeaderFieldSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.HeaderField(colander.String())

    def test_decode_unicode(self):
        value = '\xe7 is not a c'
        deserialized = self.schema.deserialize(value.encode('utf-8'))
        self.assertEquals(deserialized, value)

    def test_bad_unicode_raises_invalid(self):
        value = b'utf8 \xe9'
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)


class QueryFieldSchemaTest(unittest.TestCase):

    def test_deserialize_integer_between_quotes(self):
        self.schema = schema.QueryField(colander.Integer())
        deserialized = self.schema.deserialize("123")
        self.assertEquals(deserialized, 123)


class FieldListSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.FieldList()

    def test_deserialize_as_list(self):
        value = 'foo,-bar,123'
        deserialized = self.schema.deserialize(value)
        self.assertEquals(deserialized, ['foo', '-bar', '123'])

    def test_handle_drop(self):
        value = colander.null
        deserialized = self.schema.deserialize(value)
        self.assertEquals(deserialized, colander.drop)

    def test_handle_empty(self):
        value = ''
        deserialized = self.schema.deserialize(value)
        self.assertEquals(deserialized, [])

    def test_raises_invalid_if_not_string(self):
        value = 123
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          value)


class HeaderQuotedIntegerSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.HeaderQuotedInteger()

    def test_deserialize_as_integer(self):
        value = '"123"'
        deserialized = self.schema.deserialize(value)
        self.assertEquals(deserialized, 123)

    def test_deserialize_any(self):
        value = '*'
        deserialized = self.schema.deserialize(value)
        self.assertEquals(deserialized, '*')

    def test_unquoted_raises_invalid(self):
        value = '123'
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          value)

    def test_integer_raises_invalid(self):
        value = 123
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          value)

    def test_invalid_quoted_raises_invalid(self):
        value = 'foo'
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          value)

    def test_empty_raises_invalid(self):
        value = '""'
        self.assertRaises(colander.Invalid,
                          self.schema.deserialize,
                          value)


class RecordSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.RecordSchema()

    def test_binds_data(self):
        bound = self.schema.bind(data=schema.ResourceSchema())
        value = {'data': {'foo': 'bar'}}
        deserialized = bound.deserialize(value)
        self.assertEquals(deserialized, value)

    def test_binds_permissions(self):
        permissions = schema.PermissionsSchema(permissions=('sleep', ))
        bound = self.schema.bind(permissions=permissions)
        value = {'permissions': {'sleep': []}}
        deserialized = bound.deserialize(value)
        self.assertEquals(deserialized, value)

    def test_allow_binding_perms_after_data(self):
        bound = self.schema.bind(data=schema.ResourceSchema())
        permissions = schema.PermissionsSchema(permissions=('sleep', ))
        bound = bound.bind(permissions=permissions)
        value = {'data': {'foo': 'bar'}, 'permissions': {'sleep': []}}
        deserialized = bound.deserialize(value)
        self.assertEquals(deserialized, value)

    def test_doesnt_allow_permissions_unless_bound(self):
        bound = self.schema.bind(data=schema.ResourceSchema())
        value = {'permissions': {'sleep': []}}
        self.assertRaises(colander.UnsupportedFields, bound.deserialize, value)


class RequestSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.RequestSchema()

    def test_header_supports_binding(self):
        header = colander.MappingSchema(missing={'foo': 'bar'})
        bound = self.schema.bind(header=header)
        deserialized = bound.deserialize({})
        self.assertEquals(deserialized['header'], {'foo': 'bar'})

    def test_querystring_supports_binding(self):
        querystring = colander.MappingSchema(missing={'foo': 'bar'})
        bound = self.schema.bind(querystring=querystring)
        deserialized = bound.deserialize({})
        self.assertEquals(deserialized['querystring'], {'foo': 'bar'})

    def test_default_header_if_not_bound(self):
        bound = self.schema.bind()
        self.assertEquals(type(bound['header']), schema.HeaderSchema)

    def test_default_querystring_if_not_bound(self):
        bound = self.schema.bind()
        self.assertEquals(type(bound['querystring']), schema.QuerySchema)

    def test_header_preserve_unkown_fields(self):
        value = {'header': {'foo': 'bar'}}
        deserialized = self.schema.bind().deserialize(value)
        self.assertEquals(deserialized, value)

    def test_querystring_preserve_unkown_fields(self):
        value = {'querystring': {'foo': 'bar'}}
        deserialized = self.schema.bind().deserialize(value)
        self.assertEquals(deserialized, value)

    def test_drops(self):
        deserialized = self.schema.bind().deserialize({})
        self.assertEquals(deserialized, {})


class PayloadRequestSchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.PayloadRequestSchema()

    def test_body_supports_binding(self):
        body = colander.MappingSchema(missing={'foo': 'bar'})
        bound = self.schema.bind(body=body)
        deserialized = bound.deserialize({})
        self.assertEquals(deserialized['body'], {'foo': 'bar'})

    def test_body_supports_binding_after_other_binds(self):
        querystring = colander.MappingSchema(missing={'foo': 'bar'})
        bound = self.schema.bind(querystring=querystring)
        body = colander.MappingSchema(missing={'foo': 'beer'})
        bound = bound.bind(body=body)
        deserialized = bound.deserialize({})
        self.assertEquals(deserialized['querystring'], {'foo': 'bar'})
        self.assertEquals(deserialized['body'], {'foo': 'beer'})


class CollectionQuerySchemaTest(unittest.TestCase):

    def setUp(self):
        self.schema = schema.CollectionQuerySchema()
        self.querystring = {
            '_limit': '2',
            '_sort': 'toto,tata',
            '_token': 'abc',
            '_since': '1234',
            '_to': '7890',
            '_before': '4567',
            'id': 'toot',
            'last_modified': '9874'
        }

    def test_decode_valid_querystring(self):
        deserialized = self.schema.deserialize(self.querystring)
        self.assertEquals(deserialized, {
            '_limit': 2,
            '_sort': ['toto', 'tata'],
            '_token': 'abc',
            '_since': 1234,
            '_to': 7890,
            '_before': 4567,
            'id': 'toot',
            'last_modified': 9874
        })

    def test_raises_invalid_for_to_big_integer_in_limit(self):
        querystring = self.querystring.copy()
        querystring['_limit'] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_since(self):
        querystring = self.querystring.copy()
        querystring['_since'] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_to(self):
        querystring = self.querystring.copy()
        querystring['_to'] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_before(self):
        querystring = self.querystring.copy()
        querystring['_before'] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_last_modified(self):
        querystring = self.querystring.copy()
        querystring['last_modified'] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_limit(self):
        querystring = self.querystring.copy()
        querystring['_limit'] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_since(self):
        querystring = self.querystring.copy()
        querystring['_since'] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_to(self):
        querystring = self.querystring.copy()
        querystring['_to'] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_before(self):
        querystring = self.querystring.copy()
        querystring['_before'] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_last_modified(self):
        querystring = self.querystring.copy()
        querystring['last_modified'] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)


class ResourceReponsesTest(unittest.TestCase):

    def setUp(self):
        self.handler = schema.ResourceReponses()
        self.resource = colander.MappingSchema(title='fake')
        self.record = schema.RecordSchema().bind(data=self.resource)

    def test_get_and_bind_assign_resource_schema_to_records(self):
        responses = self.handler.get_and_bind('record', 'get',
                                              record=self.record)
        ok_response = responses['200']
        self.assertEquals(self.record['data'], ok_response['body']['data'])

    def test_get_and_bind_assign_resource_schema_to_collections(self):
        responses = self.handler.get_and_bind('collection', 'get',
                                              record=self.record)
        ok_response = responses['200']
        # XXX: Data is repeated because it's a colander sequence type index
        self.assertEquals(self.record['data'],
                          ok_response['body']['data']['data'])

    def test_responses_doesnt_have_permissions_if_not_bound(self):
        responses = self.handler.get_and_bind('record', 'get',
                                              record=self.record)
        ok_response = responses['200']
        self.assertNotIn('permissions', ok_response['body'])


class ShareableResourceReponsesTest(unittest.TestCase):

    def setUp(self):
        self.handler = schema.ShareableResourseResponses()
        self.resource = colander.MappingSchema(title='fake')
        self.permissions = colander.MappingSchema(title='bla')
        self.record = schema.RecordSchema().bind(data=self.resource,
                                                 permissions=self.permissions)

    def test_shareable_responses_doesnt_update_resource_responses(self):
        resource_handler = schema.ResourceReponses()
        shareable_responses = self.handler.get_and_bind('record', 'get')
        resource_responses = resource_handler.get_and_bind('record', 'get')
        self.assertIn('401', shareable_responses)
        self.assertNotIn('401', resource_responses)

    def test_get_and_bind_assign_permission_schema_to_records(self):
        responses = self.handler.get_and_bind('record', 'get',
                                              record=self.record)
        ok_response = responses['200']
        self.assertEquals(self.record['permissions'],
                          ok_response['body']['permissions'])



import mock
import unittest
from uuid import UUID

from pyramid.httpexceptions import HTTPBadRequest

from kinto.core.errors import ERRORS, http_error
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.testing import get_user_headers, FormattedErrorMixin
from kinto.core.utils import hmac_digest

from ..support import BaseWebTest, MINIMALIST_RECORD


class DefaultBucketWebTest(BaseWebTest, unittest.TestCase):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['includes'] = 'kinto.plugins.default_bucket'
        return settings


class DefaultBucketViewTest(FormattedErrorMixin, DefaultBucketWebTest):

    bucket_url = '/buckets/default'
    collection_url = '/buckets/default/collections/tasks'

    def test_default_bucket_exists_and_has_user_id(self):
        bucket = self.app.get(self.bucket_url, headers=self.headers)
        result = bucket.json
        settings = self.app.app.registry.settings
        hmac_secret = settings['userid_hmac_secret']
        bucket_id = hmac_digest(hmac_secret, self.principal)[:32]

        self.assertEqual(result['data']['id'], str(UUID(bucket_id)))
        self.assertEqual(result['permissions']['write'], [self.principal])

    def test_default_bucket_can_still_be_explicitly_created(self):
        bucket = {'permissions': {'read': ['system.Everyone']}}
        resp = self.app.put_json(self.bucket_url, bucket, headers=self.headers)
        result = resp.json
        self.assertIn('system.Everyone', result['permissions']['read'])

    def test_default_bucket_collection_can_still_be_explicitly_created(self):
        collection = {'data': {'synced': True}}
        resp = self.app.put_json(self.collection_url, collection, headers=self.headers)
        result = resp.json
        self.assertIn('synced', result['data'])
        self.assertTrue(result['data']['synced'])
        resp = self.app.get(self.collection_url, headers=self.headers)
        result = resp.json
        self.assertIn('synced', result['data'])
        self.assertTrue('synced', result['data']['synced'])

    def test_default_bucket_can_be_created_with_simple_put(self):
        self.app.put(self.bucket_url, headers=get_user_headers('bob'), status=201)

    def test_default_bucket_collections_are_automatically_created(self):
        self.app.get(self.collection_url, headers=self.headers, status=200)

    def test_adding_a_task_for_bob_doesnt_add_it_for_alice(self):
        record = {**MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url + '/records',
                                  record, headers=get_user_headers('bob'))
        record_id = '{}/records/{}'.format(self.collection_url, resp.json['data']['id'])
        resp = self.app.get(record_id, headers=get_user_headers('alice'),
                            status=404)

    def test_unauthenticated_bucket_access_raises_json_401(self):
        resp = self.app.get(self.bucket_url, status=401)
        self.assertEquals(resp.json['message'],
                          'Please authenticate yourself to use this endpoint.')

    def test_bucket_id_is_an_uuid_with_dashes(self):
        bucket = self.app.get(self.bucket_url, headers=self.headers)
        bucket_id = bucket.json['data']['id']
        self.assertIn('-', bucket_id)
        try:
            UUID(bucket_id)
        except ValueError:
            self.fail('bucket_id: {} is not a valid UUID.'.format(bucket_id))

    def test_second_call_on_default_bucket_doesnt_raise_a_412(self):
        self.app.get(self.bucket_url, headers=self.headers)
        self.app.get(self.bucket_url, headers=self.headers)

    def test_second_call_on_default_bucket_collection_doesnt_raise_a_412(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.headers)

    def test_querystring_parameters_are_taken_into_account(self):
        self.app.get(self.collection_url + '/records?_since=invalid',
                     headers=self.headers,
                     status=400)

    def test_option_is_possible_without_authentication_for_default(self):
        headers = 'authorization,content-type'
        self.app.options(self.collection_url + '/records',
                         headers={
                             'Origin': 'http://localhost:8000',
                             'Access-Control-Request-Method': 'GET',
                             'Access-Control-Request-Headers': headers})

    def test_cors_headers_are_provided_on_errors(self):
        resp = self.app.post_json(self.collection_url + '/records',
                                  MINIMALIST_RECORD,
                                  headers=self.headers)
        current = resp.json['data']['last_modified']
        headers = {**self.headers, 'Origin': 'http://localhost:8000',
                   'If-None-Match': ('"{}"'.format(current)).encode('utf-8')}
        resp = self.app.get(self.collection_url + '/records',
                            headers=headers, status=304)
        self.assertIn('Access-Control-Allow-Origin', resp.headers)
        self.assertIn('ETag', resp.headers['Access-Control-Expose-Headers'])

    def test_etag_is_present_and_exposed_in_304_error(self):
        resp = self.app.post_json(self.collection_url + '/records',
                                  MINIMALIST_RECORD,
                                  headers=self.headers)
        current = resp.json['data']['last_modified']
        headers = {**self.headers, 'Origin': 'http://localhost:8000',
                   'If-None-Match': ('"{}"'.format(current)).encode('utf-8')}
        resp = self.app.get(self.collection_url + '/records',
                            headers=headers, status=304)
        self.assertIn('Access-Control-Expose-Headers', resp.headers)
        self.assertIn('ETag', resp.headers)
        self.assertIn('ETag', resp.headers['Access-Control-Expose-Headers'])

    def test_bucket_id_starting_with_default_can_still_be_created(self):
        # We need to create the bucket first since it is not the default bucket
        resp = self.app.put(
            self.bucket_url.replace('default', 'default-1234'),
            headers=self.headers, status=201)
        bucket_id = resp.json['data']['id']
        self.assertEquals(bucket_id, 'default-1234')

        # We can then create the collection
        collection_url = '/buckets/default-1234/collections/default'
        self.app.put(
            collection_url,
            headers=self.headers,
            status=201)
        resp = self.app.get('/buckets/default-1234/collections',
                            headers=self.headers)
        self.assertEquals(resp.json['data'][0]['id'], 'default')

    def test_bucket_id_in_body_can_be_default(self):
        self.app.put_json('/buckets/default',
                          {'data': {'id': 'default'}},
                          headers=self.headers,
                          status=201)

    def test_default_bucket_objects_are_checked_only_once_in_batch(self):
        batch = {'requests': []}
        nb_create = 25
        for i in range(nb_create):
            request = {'method': 'POST',
                       'path': self.collection_url + '/records',
                       'body': MINIMALIST_RECORD}
            batch['requests'].append(request)

        with mock.patch.object(self.storage, 'create',
                               wraps=self.storage.create) as patched:
            self.app.post_json('/batch', batch, headers=self.headers)
            self.assertEqual(patched.call_count, nb_create + 2)

    def test_parent_collection_is_taken_from_the_one_created_in_batch(self):
        batch = {'requests': []}
        nb_create = 25
        for i in range(nb_create):
            request = {'method': 'POST',
                       'path': self.collection_url + '/records',
                       'body': MINIMALIST_RECORD}
            batch['requests'].append(request)

        with mock.patch.object(self.storage, 'create',
                               wraps=self.storage.create) as patched:
            self.app.post_json('/batch', batch, headers=self.headers)
            # Called twice only: bucket + collection ids unicity.
            self.assertEqual(patched.call_count, 25 + 2)

    def test_parent_collection_is_taken_from_the_one_checked_in_batch(self):
        # Create it first.
        self.app.put(self.collection_url, headers=self.headers, status=201)

        batch = {'requests': []}
        nb_create = 25
        for i in range(nb_create):
            request = {'method': 'POST',
                       'path': self.collection_url + '/records',
                       'body': MINIMALIST_RECORD}
            batch['requests'].append(request)

        with mock.patch.object(self.storage, 'create',
                               wraps=self.storage.create) as patched:
            self.app.post_json('/batch', batch, headers=self.headers)
            # Called twice only: bucket + collection ids unicity.
            self.assertEqual(patched.call_count, 25 + 2)

    def test_collection_id_is_validated(self):
        collection_url = '/buckets/default/collections/__files__/records'
        self.app.get(collection_url, headers=self.headers, status=400)

    def test_collection_id_does_not_support_unicode(self):
        collection_url = '/buckets/default/collections/%E8%A6%8B/records'
        self.app.get(collection_url, headers=self.headers, status=400)

    def test_405_is_a_valid_formatted_error(self):
        response = self.app.post(self.collection_url,
                                 headers=self.headers, status=405)
        self.assertFormattedError(
            response, 405, ERRORS.METHOD_NOT_ALLOWED, "Method Not Allowed",
            "Method not allowed on this endpoint.")

    def test_formatted_error_are_passed_through(self):
        response = http_error(HTTPBadRequest(),
                              errno=ERRORS.INVALID_PARAMETERS,
                              message='Yop')

        with mock.patch.object(self.storage, 'create') as mocked:
            mocked.side_effect = [
                {"id": "abc", "last_modified": 43},
                {"id": "abc", "last_modified": 44},
                response
            ]
            resp = self.app.post(self.collection_url + '/records',
                                 headers=self.headers,
                                 status=400)
            self.assertEqual(resp.body, response.body)

    def test_trailing_slash_redirection_works_for_default_bucket(self):
        collection_url = '/buckets/default/'
        resp = self.app.get(collection_url, headers=self.headers, status=307)
        assert resp.headers['Location'].endswith('/buckets/default')

    def test_trailing_slash_redirection_works_for_collections(self):
        collection_url = '/buckets/default/collections/'
        resp = self.app.get(collection_url, headers=self.headers, status=307)
        assert resp.headers['Location'].endswith('/buckets/default/collections')

    def test_trailing_slash_redirection_works_for_collection(self):
        collection_url = '/buckets/default/collections/foo/'
        resp = self.app.get(collection_url, headers=self.headers, status=307)
        assert resp.headers['Location'].endswith('/buckets/default/collections/foo')

    def test_trailing_slash_redirection_works_for_records(self):
        records_url = '/buckets/default/collections/foo/records/'
        resp = self.app.get(records_url, headers=self.headers, status=307)
        assert resp.headers['Location'].endswith('/buckets/default/collections/foo/records')

    def test_trailing_slash_redirection_works_for_record(self):
        records_url = '/buckets/default/collections/foo/records/bar/'
        resp = self.app.get(records_url, headers=self.headers, status=307)
        assert resp.headers['Location'].endswith('/buckets/default/collections/foo/records/bar')


class HelloViewTest(DefaultBucketWebTest):

    def test_returns_bucket_id_and_url_if_authenticated(self):
        response = self.app.get('/', headers=self.headers)
        self.assertEqual(response.json['user']['bucket'],
                         'ddaf8694-fa9e-2949-ed0e-77198a7907fb')

    def test_flush_capability_if_enabled(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('default_bucket', capabilities)


_events = []


def load_from_config(config, prefix):
    def listener(event):
        _events.append(event)

    return listener


class EventsTest(DefaultBucketWebTest):
    def tearDown(self):
        super().tearDown()
        del _events[:]

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings = {**settings, 'event_listeners': 'testevent',
                    'event_listeners.testevent.use': 'tests.plugins.test_default_bucket'}
        return settings

    def test_an_event_is_sent_on_implicit_bucket_creation(self):
        bucket_url = '/buckets/default'
        self.app.get(bucket_url, headers=self.headers)
        assert len(_events) == 1
        assert 'last_modified' in _events[-1].impacted_records[0]['new']
        payload = _events[-1].payload
        assert payload['resource_name'] == 'bucket'
        assert payload['action'] == 'create'

    def test_an_event_is_sent_on_implicit_collection_creation(self):
        collection_url = '/buckets/default/collections/articles'
        self.app.get(collection_url, headers=self.headers)
        assert len(_events) == 2
        payload = _events[-1].payload
        assert payload['resource_name'] == 'collection'
        assert payload['action'] == 'create'

    def test_events_sent_on_bucket_and_collection_creation(self):
        records_uri = '/buckets/default/collections/articles/records/implicit'
        body = {"data": {"implicit": "creations"}}
        self.app.put_json(records_uri, body, headers=self.headers)

        assert len(_events) == 3

        # Implicit creation of bucket
        resp = self.app.get('/', headers=self.headers)
        bucket_id = resp.json['user']['bucket']
        assert 'subpath' not in _events[0].payload
        assert _events[0].payload['action'] == 'create'
        assert _events[0].payload['bucket_id'] == bucket_id
        assert _events[0].payload['uri'] == '/buckets/{}'.format(bucket_id)

        # Implicit creation of collection
        assert 'subpath' not in _events[1].payload
        assert _events[1].payload['action'] == 'create'
        assert _events[1].payload['resource_name'] == 'collection'
        assert _events[1].payload['bucket_id'] == bucket_id
        assert _events[1].payload['collection_id'] == 'articles'
        assert _events[1].payload['uri'] == '/buckets/{}/collections/articles'.format(bucket_id)

        # Creation of record
        assert _events[2].payload['action'] == 'create'
        assert _events[2].payload['resource_name'] == 'record'
        assert _events[2].payload['bucket_id'] == bucket_id
        assert _events[2].payload['collection_id'] == 'articles'
        assert _events[2].payload['uri'] == records_uri.replace('default',
                                                                bucket_id)


class ReadonlyDefaultBucket(DefaultBucketWebTest):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['readonly'] = True
        return settings

    def test_implicit_creation_is_rejected(self):
        self.app.get('/buckets/default', headers=self.headers, status=405)


class BackendErrorTest(DefaultBucketWebTest):
    def setUp(self):
        super().setUp()
        self.patcher = mock.patch.object(
            self.storage, 'create',
            side_effect=storage_exceptions.BackendError())
        self.addCleanup(self.patcher.stop)

    def test_implicit_bucket_creation_raises_503_if_backend_fails(self):
        self.patcher.start()
        self.app.get('/buckets/default', headers=self.headers, status=503)

    def test_implicit_collection_creation_raises_503_if_backend_fails(self):
        self.app.get('/buckets/default', headers=self.headers)
        self.patcher.start()
        self.app.get('/buckets/default/collections/articles',
                     headers=self.headers, status=503)



import copy

from pyramid.httpexceptions import HTTPInsufficientStorage
from kinto.core.errors import http_error, ERRORS
from kinto.core.storage.exceptions import RecordNotFoundError
from kinto.core.utils import instance_uri

from .utils import record_size

QUOTA_RESOURCE_NAME = 'quota'
QUOTA_BUCKET_ID = 'bucket_info'
QUOTA_COLLECTION_ID = 'collection_info'


def get_bucket_settings(settings, bucket_id, name):
    return settings.get(
        # Bucket specific
        'quotas.bucket_{}_{}'.format(bucket_id, name),
        # Global to all buckets
        settings.get('quotas.bucket_{}'.format(name), None))


def get_collection_settings(settings, bucket_id, collection_id, name):
    return settings.get(
        # Specific for a given bucket collection
        'quotas.collection_{}_{}_{}'.format(bucket_id, collection_id, name),
        # Specific to given bucket collections
        settings.get('quotas.collection_{}_{}'.format(bucket_id, name),
                     # Global to all buckets collections
                     settings.get('quotas.collection_{}'.format(name), None)))


def on_resource_changed(event):
    """
    Everytime an object is created/changed/deleted, we update the
    bucket counters.

    If a new object exceeds the quotas, we reject the request.
    """
    payload = event.payload
    action = payload['action']
    resource_name = payload['resource_name']
    event_uri = payload['uri']

    settings = event.request.registry.settings

    bucket_id = payload['bucket_id']
    bucket_uri = instance_uri(event.request, 'bucket', id=bucket_id)
    collection_id = None
    collection_uri = None
    if 'collection_id' in payload:
        collection_id = payload['collection_id']
        collection_uri = instance_uri(event.request,
                                      'collection',
                                      bucket_id=bucket_id,
                                      id=collection_id)

    bucket_max_bytes = get_bucket_settings(settings, bucket_id, 'max_bytes')
    bucket_max_items = get_bucket_settings(settings, bucket_id, 'max_items')
    bucket_max_bytes_per_item = get_bucket_settings(settings, bucket_id,
                                                    'max_bytes_per_item')
    collection_max_bytes = get_collection_settings(settings, bucket_id,
                                                   collection_id, 'max_bytes')
    collection_max_items = get_collection_settings(settings, bucket_id,
                                                   collection_id, 'max_items')
    collection_max_bytes_per_item = get_collection_settings(
        settings, bucket_id, collection_id, 'max_bytes_per_item')

    max_bytes_per_item = (collection_max_bytes_per_item or
                          bucket_max_bytes_per_item)

    storage = event.request.registry.storage

    if action == 'delete' and resource_name == 'bucket':
        try:
            storage.delete(parent_id=bucket_uri,
                           collection_id=QUOTA_RESOURCE_NAME,
                           object_id=QUOTA_BUCKET_ID)
        except RecordNotFoundError:
            pass

        collection_pattern = instance_uri(event.request, 'collection',
                                          bucket_id=bucket_id, id='*')
        storage.delete_all(parent_id=collection_pattern,
                           collection_id=QUOTA_RESOURCE_NAME)
        return

    targets = []
    for impacted in event.impacted_records:
        target = impacted['new' if action != 'delete' else 'old']
        # On POST .../records, the URI does not contain the newly created
        # record id.
        obj_id = target['id']
        parts = event_uri.split('/')
        if resource_name in parts[-1]:
            parts.append(obj_id)
        else:
            # Make sure the id is correct on grouped events.
            parts[-1] = obj_id
        uri = '/'.join(parts)

        old = impacted.get('old', {})
        new = impacted.get('new', {})

        targets.append((uri, obj_id, old, new))

    try:
        bucket_info = copy.deepcopy(
            storage.get(parent_id=bucket_uri,
                        collection_id=QUOTA_RESOURCE_NAME,
                        object_id=QUOTA_BUCKET_ID))
    except RecordNotFoundError:
        bucket_info = {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": 0,
        }

    collection_info = {
        "record_count": 0,
        "storage_size": 0,
    }
    if collection_id:
        try:
            collection_info = copy.deepcopy(
                storage.get(parent_id=collection_uri,
                            collection_id=QUOTA_RESOURCE_NAME,
                            object_id=QUOTA_COLLECTION_ID))
        except RecordNotFoundError:
            pass

    # Update the bucket quotas values for each impacted record.
    for (uri, obj_id, old, new) in targets:
        old_size = record_size(old)
        new_size = record_size(new)

        if max_bytes_per_item is not None and action != "delete":
            if new_size > max_bytes_per_item:
                message = ("Maximum bytes per object exceeded "
                           "({} > {} Bytes.".format(new_size, max_bytes_per_item))
                raise http_error(HTTPInsufficientStorage(),
                                 errno=ERRORS.FORBIDDEN.value,
                                 message=message)

        if action == 'create':
            bucket_info['storage_size'] += new_size
            if resource_name == 'collection':
                bucket_info['collection_count'] += 1
                collection_info['storage_size'] += new_size
            if resource_name == 'record':
                bucket_info['record_count'] += 1
                collection_info['record_count'] += 1
                collection_info['storage_size'] += new_size
        elif action == 'update':
            bucket_info['storage_size'] -= old_size
            bucket_info['storage_size'] += new_size
            if resource_name in ('collection', 'record'):
                collection_info['storage_size'] -= old_size
                collection_info['storage_size'] += new_size
        else:   # action == 'delete':
            bucket_info['storage_size'] -= old_size
            if resource_name == 'collection':
                collection_uri = uri
                bucket_info['collection_count'] -= 1
                # When we delete the collection all the records in it
                # are deleted without notification.
                collection_records, _ = storage.get_all(
                    collection_id='record',
                    parent_id=collection_uri)
                for r in collection_records:
                    old_record_size = record_size(r)
                    bucket_info['record_count'] -= 1
                    bucket_info['storage_size'] -= old_record_size
                    collection_info['record_count'] -= 1
                    collection_info['storage_size'] -= old_record_size
                collection_info['storage_size'] -= old_size

            if resource_name == 'record':
                bucket_info['record_count'] -= 1
                collection_info['record_count'] -= 1
                collection_info['storage_size'] -= old_size

    if bucket_max_bytes is not None:
        if bucket_info['storage_size'] > bucket_max_bytes:
            message = ("Bucket maximum total size exceeded "
                       "({} > {} Bytes). ".format(bucket_info['storage_size'],
                                                  bucket_max_bytes))
            raise http_error(HTTPInsufficientStorage(),
                             errno=ERRORS.FORBIDDEN.value,
                             message=message)

    if bucket_max_items is not None:
        if bucket_info['record_count'] > bucket_max_items:
            message = ("Bucket maximum number of objects exceeded "
                       "({} > {} objects).".format(bucket_info['record_count'],
                                                   bucket_max_items))
            raise http_error(HTTPInsufficientStorage(),
                             errno=ERRORS.FORBIDDEN.value,
                             message=message)

    if collection_max_bytes is not None:
        if collection_info['storage_size'] > collection_max_bytes:
            message = ("Collection maximum size exceeded "
                       "({} > {} Bytes).".format(collection_info['storage_size'],
                                                 collection_max_bytes))
            raise http_error(HTTPInsufficientStorage(),
                             errno=ERRORS.FORBIDDEN.value,
                             message=message)

    if collection_max_items is not None:
        if collection_info['record_count'] > collection_max_items:
            message = ("Collection maximum number of objects exceeded "
                       "({} > {} objects).".format(collection_info['record_count'],
                                                   collection_max_items))
            raise http_error(HTTPInsufficientStorage(),
                             errno=ERRORS.FORBIDDEN.value,
                             message=message)

    storage.update(parent_id=bucket_uri,
                   collection_id=QUOTA_RESOURCE_NAME,
                   object_id=QUOTA_BUCKET_ID,
                   record=bucket_info)

    if collection_id:
        if action == 'delete' and resource_name == 'collection':
            try:
                storage.delete(parent_id=collection_uri,
                               collection_id=QUOTA_RESOURCE_NAME,
                               object_id=QUOTA_COLLECTION_ID)
            except RecordNotFoundError:
                pass
            return
        else:
            storage.update(parent_id=collection_uri,
                           collection_id=QUOTA_RESOURCE_NAME,
                           object_id=QUOTA_COLLECTION_ID,
                           record=collection_info)



class Model:
    """A collection stores and manipulate records in its attached storage.

    It is not aware of HTTP environment nor HTTP API.

    Records are isolated according to the provided `name` and `parent_id`.

    Those notions have no particular semantic and can represent anything.
    For example, the collection `name` can be the *type* of objects stored, and
    `parent_id` can be the current *user id* or *a group* where the collection
    belongs. If left empty, the collection records are not isolated.
    """
    id_field = 'id'
    """Name of `id` field in records"""

    modified_field = 'last_modified'
    """Name of `last modified` field in records"""

    deleted_field = 'deleted'
    """Name of `deleted` field in deleted records"""

    def __init__(self, storage, id_generator=None, collection_id='',
                 parent_id='', auth=None):
        """
        :param storage: an instance of storage
        :type storage: :class:`kinto.core.storage.Storage`
        :param id_generator: an instance of id generator, used by storage
            on record creation.

        :param str collection_id: the collection id
        :param str parent_id: the default parent id
        """
        self.storage = storage
        self.id_generator = id_generator
        self.parent_id = parent_id
        self.collection_id = collection_id
        self.auth = auth

    def timestamp(self, parent_id=None):
        """Fetch the collection current timestamp.

        :param str parent_id: optional filter for parent id
        :rtype: int
        """
        parent_id = parent_id or self.parent_id
        return self.storage.collection_timestamp(
            collection_id=self.collection_id,
            parent_id=parent_id,
            auth=self.auth)

    def get_records(self, filters=None, sorting=None, pagination_rules=None,
                    limit=None, include_deleted=False, parent_id=None):
        """Fetch the collection records.

        Override to post-process records after feching them from storage.

        :param filters: Optionally filter the records by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`

        :param sorting: Optionnally sort the records by attribute.
            Each sort instruction in this list refers to a field and a
            direction (negative means descending). All sort instructions are
            cumulative.
        :type sorting: list of :class:`kinto.core.storage.Sort`

        :param pagination_rules: Optionnally paginate the list of records.
            This list of rules aims to reduce the set of records to the current
            page. A rule is a list of filters (see `filters` parameter),
            and all rules are combined using *OR*.
        :type pagination_rules: list of list of
            :class:`kinto.core.storage.Filter`

        :param int limit: Optionnally limit the number of records to be
            retrieved.

        :param bool include_deleted: Optionnally include the deleted records
            that match the filters.

        :param str parent_id: optional filter for parent id

        :returns: A tuple with the list of records in the current page,
            the total number of records in the result set.
        :rtype: tuple
        """
        parent_id = parent_id or self.parent_id
        records, total_records = self.storage.get_all(
            collection_id=self.collection_id,
            parent_id=parent_id,
            filters=filters,
            sorting=sorting,
            pagination_rules=pagination_rules,
            limit=limit,
            include_deleted=include_deleted,
            id_field=self.id_field,
            modified_field=self.modified_field,
            deleted_field=self.deleted_field,
            auth=self.auth)
        return records, total_records

    def delete_records(self, filters=None, sorting=None, pagination_rules=None,
                       limit=None, parent_id=None):
        """Delete multiple collection records.

        Override to post-process records after their deletion from storage.

        :param filters: Optionally filter the records by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`

        :param sorting: Optionnally sort the records by attribute.
            Each sort instruction in this list refers to a field and a
            direction (negative means descending). All sort instructions are
            cumulative.
        :type sorting: list of :class:`kinto.core.storage.Sort`

        :param pagination_rules: Optionnally paginate the deletion of records.
            This list of rules aims to reduce the set of records to the current
            page. A rule is a list of filters (see `filters` parameter),
            and all rules are combined using *OR*.
        :type pagination_rules: list of list of
            :class:`kinto.core.storage.Filter`

        :param int limit: Optionnally limit the number of records to be
           deleted.

        :param str parent_id: optional filter for parent id

        :returns: The list of deleted records from storage.
        """
        parent_id = parent_id or self.parent_id
        return self.storage.delete_all(collection_id=self.collection_id,
                                       parent_id=parent_id,
                                       filters=filters,
                                       sorting=sorting,
                                       pagination_rules=pagination_rules,
                                       limit=limit,
                                       id_field=self.id_field,
                                       modified_field=self.modified_field,
                                       deleted_field=self.deleted_field,
                                       auth=self.auth)

    def get_record(self, record_id, parent_id=None):
        """Fetch current view related record, and raise 404 if missing.

        :param str record_id: record identifier
        :param str parent_id: optional filter for parent id

        :returns: the record from storage
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        return self.storage.get(collection_id=self.collection_id,
                                parent_id=parent_id,
                                object_id=record_id,
                                id_field=self.id_field,
                                modified_field=self.modified_field,
                                auth=self.auth)

    def create_record(self, record, parent_id=None, ignore_conflict=False):
        """Create a record in the collection.

        Override to perform actions or post-process records after their
        creation in storage.

        .. code-block:: python

            def create_record(self, record):
                record = super().create_record(record)
                idx = index.store(record)
                record['index'] = idx
                return record

        :param dict record: record to store
        :param str parent_id: optional filter for parent id

        :returns: the newly created record.
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        return self.storage.create(collection_id=self.collection_id,
                                   parent_id=parent_id,
                                   record=record,
                                   id_generator=self.id_generator,
                                   id_field=self.id_field,
                                   modified_field=self.modified_field,
                                   auth=self.auth,
                                   ignore_conflict=ignore_conflict)

    def update_record(self, record, parent_id=None):
        """Update a record in the collection.

        Override to perform actions or post-process records after their
        modification in storage.

        .. code-block:: python

            def update_record(self, record, parent_id=None):
                record = super().update_record(record, parent_id)
                subject = 'Record {} was changed'.format(record[self.id_field])
                send_email(subject)
                return record

        :param dict record: record to store
        :param str parent_id: optional filter for parent id
        :returns: the updated record.
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        record_id = record[self.id_field]
        return self.storage.update(collection_id=self.collection_id,
                                   parent_id=parent_id,
                                   object_id=record_id,
                                   record=record,
                                   id_field=self.id_field,
                                   modified_field=self.modified_field,
                                   auth=self.auth)

    def delete_record(self, record, parent_id=None, last_modified=None):
        """Delete a record in the collection.

        Override to perform actions or post-process records after deletion
        from storage for example:

        .. code-block:: python

            def delete_record(self, record):
                deleted = super().delete_record(record)
                erase_media(record)
                deleted['media'] = 0
                return deleted

        :param dict record: the record to delete
        :param dict record: record to store
        :param str parent_id: optional filter for parent id
        :returns: the deleted record.
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        record_id = record[self.id_field]
        return self.storage.delete(collection_id=self.collection_id,
                                   parent_id=parent_id,
                                   object_id=record_id,
                                   id_field=self.id_field,
                                   modified_field=self.modified_field,
                                   deleted_field=self.deleted_field,
                                   auth=self.auth,
                                   last_modified=last_modified)


class ShareableModel(Model):
    """A protected collection interacts with the permission backend.
    """
    permissions_field = '__permissions__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Permission backend.
        self.permission = None
        # Object permission id.
        self.get_permission_object_id = None
        # Current user main principal.
        self.current_principal = None
        self.prefixed_principals = None

    def _allow_write(self, perm_object_id):
        """Helper to give the ``write`` permission to the current user.
        """
        self.permission.add_principal_to_ace(perm_object_id,
                                             'write',
                                             self.current_principal)

    def _annotate(self, record, perm_object_id):
        permissions = self.permission.get_object_permissions(perm_object_id)
        # Permissions are not returned if user only has read permission.
        writers = permissions.get('write', [])
        principals = self.prefixed_principals + [self.current_principal]
        if len(set(writers) & set(principals)) == 0:
            permissions = {}
        # Insert the permissions values in the response.
        annotated = {**record, self.permissions_field: permissions}
        return annotated

    def delete_records(self, filters=None, sorting=None, pagination_rules=None,
                       limit=None, parent_id=None):
        """Delete permissions when collection records are deleted in bulk.
        """
        deleted = super().delete_records(filters, sorting, pagination_rules, limit, parent_id)
        # Take a huge shortcut in case we want to delete everything.
        if not filters:
            perm_ids = [self.get_permission_object_id(object_id='*')]
        else:
            perm_ids = [self.get_permission_object_id(object_id=r[self.id_field])
                        for r in deleted]
        self.permission.delete_object_permissions(*perm_ids)
        return deleted

    def get_record(self, record_id, parent_id=None):
        """Fetch current permissions and add them to returned record.
        """
        record = super().get_record(record_id, parent_id)
        perm_object_id = self.get_permission_object_id(record_id)

        return self._annotate(record, perm_object_id)

    def create_record(self, record, parent_id=None, ignore_conflict=False):
        """Create record and set specified permissions.

        The current principal is added to the owner (``write`` permission).
        """
        permissions = record.pop(self.permissions_field, {})
        record = super().create_record(record, parent_id, ignore_conflict=ignore_conflict)
        record_id = record[self.id_field]
        perm_object_id = self.get_permission_object_id(record_id)
        self.permission.replace_object_permissions(perm_object_id, permissions)
        self._allow_write(perm_object_id)

        return self._annotate(record, perm_object_id)

    def update_record(self, record, parent_id=None):
        """Update record and the specified permissions.

        If no permissions is specified, the current permissions are not
        modified.

        The current principal is added to the owner (``write`` permission).
        """
        permissions = record.pop(self.permissions_field, {})
        record = super().update_record(record, parent_id)
        record_id = record[self.id_field]
        perm_object_id = self.get_permission_object_id(record_id)
        self.permission.replace_object_permissions(perm_object_id, permissions)
        self._allow_write(perm_object_id)

        return self._annotate(record, perm_object_id)

    def delete_record(self, record_id, parent_id=None, last_modified=None):
        """Delete record and its associated permissions.
        """
        record = super().delete_record(
            record_id, parent_id, last_modified=last_modified)
        perm_object_id = self.get_permission_object_id(record_id)
        self.permission.delete_object_permissions(perm_object_id)

        return record



from kinto.core.testing import unittest

from kinto.authorization import (_resource_endpoint, _relative_object_uri,
                                 _inherited_permissions)


class ResourceEndpointTest(unittest.TestCase):
    def test_resource_endpoint_return_right_type_for_key(self):
        self.assertEqual(_resource_endpoint('/buckets/bid/collections/cid/records/rid'),
                         ('record', False))
        self.assertEqual(_resource_endpoint('/buckets/bid/collections/cid'),
                         ('collection', False))
        self.assertEqual(_resource_endpoint('/buckets'), ('', False))
        self.assertEqual(_resource_endpoint('/buckets/bid'), ('bucket', False))
        self.assertEqual(_resource_endpoint('/buckets/bid/history/xx'), ('history', False))
        self.assertEqual(_resource_endpoint('/buckets/bid/groups/moderators'),
                         ('group', False))

    def test_resource_endpoint_return_right_type_for_children_collection(self):
        self.assertEqual(_resource_endpoint('/buckets/bid/collections/cid/records'),
                         ('collection', True))
        self.assertEqual(_resource_endpoint('/buckets/bid/collections'), ('bucket', True))
        self.assertEqual(_resource_endpoint('/buckets/bid/history'), ('bucket', True))
        self.assertEqual(_resource_endpoint('/buckets/bid/groups'), ('bucket', True))


class RelativeObjectUri(unittest.TestCase):
    record_uri = '/buckets/blog/collections/articles/records/article1'
    records_uri = '/buckets/blog/collections/articles/records'
    collection_uri = '/buckets/blog/collections/articles'
    collections_uri = '/buckets/blog/collections'
    group_uri = '/buckets/blog/groups/moderators'
    groups_uri = '/buckets/blog/groups'
    bucket_uri = '/buckets/blog'
    buckets_uri = '/buckets'

    def test_related_object_uri_can_construct_parents_set_uris(self):
        record_uri = '/buckets/bid/collections/cid/records/rid'
        self.assertEqual(_relative_object_uri('record', record_uri), record_uri)

        self.assertEqual(_relative_object_uri('collection', self.record_uri),
                         self.collection_uri)
        self.assertEqual(_relative_object_uri('collection', self.records_uri),
                         self.collection_uri)

        # Can build bucket_uri from obj_parts
        self.assertEqual(_relative_object_uri('bucket', self.record_uri),
                         self.bucket_uri)

        # Can build group_uri from group obj_parts
        self.assertEqual(_relative_object_uri('group', self.group_uri),
                         self.group_uri)

        # Can build bucket_uri from group obj_parts
        self.assertEqual(_relative_object_uri('bucket', self.group_uri),
                         self.bucket_uri)
        self.assertEqual(_relative_object_uri('bucket', self.groups_uri),
                         self.bucket_uri)

    def test_relative_object_uri_fail_construct_children_set_uris(self):
        # Cannot build record_uri from bucket obj_parts
        self.assertRaises(ValueError,
                          _relative_object_uri,
                          'record', self.bucket_uri)

        # Cannot build collection_uri from obj_parts
        self.assertRaises(ValueError,
                          _relative_object_uri,
                          'collection', self.bucket_uri)

        # Cannot build bucket_uri from empty obj_parts
        self.assertRaises(ValueError,
                          _relative_object_uri,
                          'collection', '')


class PermissionInheritanceTest(unittest.TestCase):
    record_uri = '/buckets/blog/collections/articles/records/article1'
    collection_uri = '/buckets/blog/collections/articles'
    group_uri = '/buckets/blog/groups/moderators'
    bucket_uri = '/buckets/blog'

    def test_relative_object_uri_fail_on_wrong_type(self):
        self.assertRaises(ValueError,
                          _relative_object_uri,
                          'schema', self.record_uri)

    def test_inherited_permissions_supports_buckets_named_collections(self):
        uri = '/buckets/collections'
        self.assertEquals(set(_inherited_permissions(uri, 'write')),
                          set([(uri, 'write')]))

    def test_inherited_permissions_for_bucket_permission(self):
        # write
        self.assertEquals(
            _inherited_permissions(self.bucket_uri, 'write'),
            [(self.bucket_uri, 'write')])
        # read
        self.assertEquals(
            set(_inherited_permissions(self.bucket_uri, 'read')),
            set([(self.bucket_uri, 'write'),
                 (self.bucket_uri, 'read'),
                 (self.bucket_uri, 'collection:create'),
                 (self.bucket_uri, 'group:create')]))

        # group:create
        groups_uri = self.bucket_uri + '/groups'
        self.assertEquals(
            set(_inherited_permissions(groups_uri, 'group:create')),
            set(
                [(self.bucket_uri, 'write'),
                 (self.bucket_uri, 'group:create')])
            )

        # collection:create
        collections_uri = self.bucket_uri + '/collections'
        self.assertEquals(
            set(_inherited_permissions(collections_uri, 'collection:create')),
            set([(self.bucket_uri, 'write'),
                 (self.bucket_uri, 'collection:create')]))

    def test_inherited_permissions_for_group_permission(self):
        # write
        self.assertEquals(
            _inherited_permissions(self.group_uri, 'write'),
            [(self.group_uri, 'write'),
             (self.bucket_uri, 'write')])
        # read
        self.assertEquals(
            set(_inherited_permissions(self.group_uri, 'read')),
            set([(self.bucket_uri, 'write'),
                 (self.bucket_uri, 'read'),
                 (self.group_uri, 'write'),
                 (self.group_uri, 'read')]))

    def test_inherited_permissions_for_collection_permission(self):
        # write
        self.assertEquals(
            _inherited_permissions(self.collection_uri, 'write'),
            [(self.collection_uri, 'write'),
             (self.bucket_uri, 'write')])
        # read
        self.assertEquals(
            set(_inherited_permissions(self.collection_uri, 'read')),
            set([(self.bucket_uri, 'write'),
                 (self.bucket_uri, 'read'),
                 (self.collection_uri, 'write'),
                 (self.collection_uri, 'read'),
                 (self.collection_uri, 'record:create')]))
        # records:create
        records_uri = self.collection_uri + '/records'
        self.assertEquals(
            set(_inherited_permissions(records_uri, 'record:create')),
            set([(self.bucket_uri, 'write'),
                 (self.collection_uri, 'write'),
                 (self.collection_uri, 'record:create')]))

    def test_inherited_permissions_for_record_permission(self):
        # write
        self.assertEquals(
            set(_inherited_permissions(self.record_uri, 'write')),
            set([(self.bucket_uri, 'write'),
                 (self.collection_uri, 'write'),
                 (self.record_uri, 'write')]))
        # read
        self.assertEquals(
            set(_inherited_permissions(self.record_uri, 'read')),
            set([(self.bucket_uri, 'write'),
                 (self.bucket_uri, 'read'),
                 (self.collection_uri, 'write'),
                 (self.collection_uri, 'read'),
                 (self.record_uri, 'write'),
                 (self.record_uri, 'read')]))

    def test_inherited_permissions_for_list_of_buckets(self):
        permissions = _inherited_permissions('/buckets', 'read')
        self.assertEquals(permissions, [])

    def test_inherited_permissions_for_non_resource_url(self):
        unknown = '/resource/unknown'
        permissions = _inherited_permissions(unknown, 'read')
        self.assertEquals(permissions, [])

    def test_inherited_permissions_for_attachment_url(self):
        attachment = '/buckets/bid/collections/cid/records/rid/attachment'
        permissions = _inherited_permissions(attachment, 'read')
        self.assertIn(('/buckets/bid/collections/cid/records/rid', 'read'), permissions)



from pyramid import httpexceptions

from kinto.core.errors import ERRORS
from . import BaseTest


class FilteringTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.validated = self.resource.request.validated
        self.patch_known_field.start()
        records = [
            {'title': 'MoFo', 'status': 0, 'favorite': True},
            {'title': 'MoFo', 'status': 1, 'favorite': False},
            {'title': 'MoFo', 'status': 2, 'favorite': False},
            {'title': 'MoFo', 'status': 0, 'favorite': False},
            {'title': 'MoFo', 'status': 1, 'favorite': True},
            {'title': 'MoFo', 'status': 2, 'favorite': False},
            {'title': 'Foo', 'status': 3, 'favorite': False},
            {'title': 'Bar', 'status': 3, 'favorite': False},
        ]
        for r in records:
            self.model.create_record(r)

    def test_list_can_be_filtered_on_deleted_with_since(self):
        since = self.model.timestamp()
        r = self.model.create_record({})
        self.model.delete_record(r)
        self.validated['querystring'] = {'_since': since, 'deleted': True}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)
        self.assertTrue(result['data'][0]['deleted'])

    def test_filter_on_id_is_supported(self):
        self.patch_known_field.stop()
        r = self.model.create_record({})
        self.validated['querystring'] = {'id': '{}'.format(r['id'])}
        result = self.resource.collection_get()
        self.assertEqual(result['data'][0], r)

    def test_list_cannot_be_filtered_on_deleted_without_since(self):
        r = self.model.create_record({})
        self.model.delete_record(r)
        self.validated['querystring'] = {'deleted': True}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_filter_works_with_empty_list(self):
        self.resource.model.parent_id = 'alice'
        self.validated['querystring'] = {'status': 1}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_number_of_records_matches_filter(self):
        self.validated['querystring'] = {'status': 1}
        self.resource.collection_get()
        headers = self.last_response.headers
        self.assertEqual(int(headers['Total-Records']), 2)

    def test_single_basic_filter_by_attribute(self):
        self.validated['querystring'] = {'status': 1}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 2)

    def test_filter_on_unknown_attribute_raises_error(self):
        self.patch_known_field.stop()
        self.validated['querystring'] = {'foo': 1}
        self.assertRaises(httpexceptions.HTTPBadRequest,
                          self.resource.collection_get)

    def test_filter_errors_are_json_formatted(self):
        self.patch_known_field.stop()
        self.validated['querystring'] = {'foo': 1}
        try:
            self.resource.collection_get()
        except httpexceptions.HTTPBadRequest as e:
            error = e
        self.assertEqual(error.json, {
            'errno': ERRORS.INVALID_PARAMETERS.value,
            'message': "Unknown filter field 'foo'",
            'code': 400,
            'error': 'Invalid parameters',
            'details': [{'description': "Unknown filter field 'foo'",
                         'location': 'querystring',
                         'name': 'foo'}]})

    def test_regexp_is_strict_for_min_and_max(self):
        self.patch_known_field.stop()
        self.validated['querystring'] = {'madmax_status': 1}
        self.assertRaises(httpexceptions.HTTPBadRequest,
                          self.resource.collection_get)

    def test_double_basic_filter_by_attribute(self):
        self.validated['querystring'] = {'status': 1, 'favorite': True}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)

    def test_string_filters_naively_by_value(self):
        self.validated['querystring'] = {'title': 'MoF'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)
        self.validated['querystring'] = {'title': 'MoFo'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 6)

    def test_string_filters_searching_by_value_not_matching(self):
        self.validated['querystring'] = {'like_title': 'MoFoo'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_string_filters_searching_by_value_matching_many(self):
        self.validated['querystring'] = {'like_title': 'Fo'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 7)

    def test_string_filters_searching_by_value_matching_one(self):
        self.validated['querystring'] = {'like_title': 'Bar'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)

    def test_string_filters_searching_by_value_matching_vary_case(self):
        self.validated['querystring'] = {'like_title': 'FoO'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)

    def test_filter_considers_string_if_syntaxically_invalid(self):
        self.validated['querystring'] = {'status': '1.2.3'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_filter_does_not_fail_with_complex_type_syntax(self):
        self.validated['querystring'] = {'status': '(1,2,3)'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_different_value(self):
        self.validated['querystring'] = {'not_status': 2}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertTrue(all([value != 2 for value in values]))

    def test_minimal_value(self):
        self.validated['querystring'] = {'min_status': 2}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertTrue(all([value >= 2 for value in values]))

    def test_gt_value(self):
        self.validated['querystring'] = {'gt_status': 2}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertTrue(all([value > 2 for value in values]))

    def test_maximal_value(self):
        self.validated['querystring'] = {'max_status': 2}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertTrue(all([value <= 2 for value in values]))

    def test_lt_value(self):
        self.validated['querystring'] = {'lt_status': 2}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertTrue(all([value < 2 for value in values]))

    def test_in_values(self):
        self.validated['querystring'] = {'in_status': [0, 1]}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertEqual(sorted(values), [0, 0, 1, 1])

    def test_exclude_values(self):
        self.validated['querystring'] = {'exclude_status': [0]}
        result = self.resource.collection_get()
        values = [item['status'] for item in result['data']]
        self.assertEqual(sorted(values), [1, 1, 2, 2, 3, 3])

    def test_include_returns_400_if_value_has_wrong_type(self):
        self.validated['querystring'] = {'in_id': [0, 1]}
        with self.assertRaises(httpexceptions.HTTPBadRequest) as cm:
            self.resource.collection_get()
        self.assertIn('in_id', cm.exception.json['message'])

        self.validated['querystring'] = {'in_last_modified': ['a', 'b']}
        self.assertRaises(httpexceptions.HTTPBadRequest,
                          self.resource.collection_get)

    def test_exclude_returns_400_if_value_has_wrong_type(self):
        self.validated['querystring'] = {'exclude_id': [0, 1]}
        with self.assertRaises(httpexceptions.HTTPBadRequest) as cm:
            self.resource.collection_get()
        self.assertIn('exclude_id', cm.exception.json['message'])

        self.validated['querystring'] = {'exclude_last_modified': ['a', 'b']}
        self.assertRaises(httpexceptions.HTTPBadRequest,
                          self.resource.collection_get)


class SubobjectFilteringTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.validated = self.resource.request.validated
        self.patch_known_field.start()
        for i in range(6):
            record = {
                'party': {'candidate': 'Marie', 'voters': i},
                'location': 'Creuse'
            }
            self.model.create_record(record)

    def test_records_can_be_filtered_by_subobjects(self):
        self.validated['querystring'] = {'party.voters': 1}
        result = self.resource.collection_get()
        values = [item['party']['voters'] for item in result['data']]
        self.assertEqual(sorted(values), [1])

    def test_subobjects_filters_are_ignored_if_not_object(self):
        self.validated['querystring'] = {'location.city': 'barcelona'}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_subobjects_filters_works_with_directives(self):
        self.validated['querystring'] = {'in_party.voters': [1, 2, 3]}
        result = self.resource.collection_get()
        values = [item['party']['voters'] for item in result['data']]
        self.assertEqual(sorted(values), [1, 2, 3])



class ListenerBase:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, event):
        """
        :param event: Incoming event
        """
        raise NotImplementedError()



import json
import mock
import time

from kinto.core.testing import ThreadMixin

from . import BaseTest


class SinceModifiedTest(ThreadMixin, BaseTest):

    def setUp(self):
        super().setUp()
        self.validated['body'] = {'data': {}}

        with mock.patch.object(self.model.storage,
                               '_bump_timestamp') as msec_mocked:
            for i in range(6):
                msec_mocked.return_value = i
                self.resource.collection_post()

    def test_filter_with_since_is_exclusive(self):
        self.validated['querystring'] = {'_since': 3}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 2)

    def test_filter_with__to_is_exclusive(self):
        self.validated['querystring'] = {'_to': 3}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 3)

    def test_filter_with__before_is_exclusive(self):
        self.validated['querystring'] = {'_before': 3}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 3)

    def test_filter_with__to_return_an_alert_header(self):
        self.validated['querystring'] = {'_to': 3}
        self.resource.collection_get()
        self.assertIn('Alert', self.resource.request.response.headers)
        alert = self.resource.request.response.headers['Alert']
        self.assertDictEqual(
            json.loads(alert),
            {
                'code': 'soft-eol',
                'message': ('_to is now deprecated, '
                            'you should use _before instead'),
                'url': ('https://kinto.readthedocs.io/en/2.4.0/api/resource'
                        '.html#list-of-available-url-parameters')
            })

    def test_get_timestamp_header_is_equal_to_last_modification(self):
        result = self.resource.collection_post()['data']
        modification = result['last_modified']
        self.resource = self.resource_class(request=self.get_request(),
                                            context=self.get_context())
        self.resource.request.validated = self.validated
        self.resource.collection_get()
        header = int(self.last_response.headers['ETag'][1:-1])
        self.assertEqual(header, modification)

    def test_delete_timestamp_header_is_equal_to_last_deleted(self):
        self.resource.collection_post()['data']
        self.resource = self.resource_class(request=self.get_request(),
                                            context=self.get_context())
        self.resource.request.validated = self.validated
        result = self.resource.collection_delete()
        modification = max([record['last_modified'] for record in result['data']])
        header = int(self.last_response.headers['ETag'][1:-1])
        self.assertEqual(header, modification)

    def test_post_timestamp_header_is_equal_to_creation(self):
        result = self.resource.collection_post()['data']
        modification = result['last_modified']
        header = int(self.last_response.headers['ETag'][1:-1])
        self.assertEqual(header, modification)

    def test_filter_with_since_accepts_numeric_value(self):
        self.validated['querystring'] = {'_since': 6}
        self.resource.collection_post()
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)

    def test_filter_from_last_modified_is_exclusive(self):
        result = self.resource.collection_post()['data']
        current = result['last_modified']

        self.validated['querystring'] = {'_since': current}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_filter_with_last_modified_includes_deleted_data(self):
        self.resource.collection_post()
        result = self.resource.collection_post()['data']
        current = result['last_modified']

        self.resource.record_id = result['id']
        self.resource.delete()

        self.validated['querystring'] = {'_since': current}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)
        self.assertTrue(result['data'][0]['deleted'])

    def test_filter_from_last_header_value_is_exclusive(self):
        self.resource.collection_get()
        current = int(self.last_response.headers['ETag'][1:-1])

        self.validated['querystring'] = {'_since': current}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_filter_works_with_empty_list(self):
        self.resource.model.parent_id = 'alice'
        self.validated['querystring'] = {'_since': 3}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_timestamp_are_always_identical_on_read(self):

        def read_timestamp():
            self.resource.collection_get()
            return int(self.last_response.headers['ETag'][1:-1])

        before = read_timestamp()
        now = read_timestamp()
        after = read_timestamp()
        self.assertEqual(before, now)
        self.assertEqual(now, after)

    def test_timestamp_are_always_incremented_on_creation(self):

        def read_timestamp():
            record = self.resource.collection_post()['data']
            return record['last_modified']

        before = read_timestamp()
        now = read_timestamp()
        after = read_timestamp()
        self.assertTrue(before < now < after)

    def test_records_created_during_fetch_are_above_fetch_timestamp(self):

        timestamps = {}

        def long_fetch():
            """Simulate a overhead while reading on storage."""

            def delayed_get(*args, **kwargs):
                time.sleep(.100)  # 100 msec
                return [], 0

            with mock.patch.object(self.model.storage,
                                   'get_all', delayed_get):
                self.resource.collection_get()
                fetch_at = self.last_response.headers['ETag'][1:-1]
                timestamps['fetch'] = int(fetch_at)

        # Create a real record with no patched timestamp
        self.resource.collection_post()

        # Some client start fetching
        thread = self._create_thread(target=long_fetch)
        thread.start()

        # Create record while other is fetching
        time.sleep(.020)  # 20 msec
        # Instantiate a new resource/request to avoid shared references with
        # the other one running in a thread:
        resource = self.resource_class(request=self.get_request(),
                                       context=self.get_context())
        resource.request.validated = self.validated
        resource.request.validated['body'] = {'data': {}}
        record = resource.collection_post()['data']
        timestamps['post'] = record['last_modified']

        # Wait for the fetch to finish
        thread.join()

        # Make sure fetch timestamp is below (for next fetch)
        self.assertGreater(timestamps['post'], timestamps['fetch'])



import mock

from kinto.core.testing import unittest

from .support import BaseWebTest


class SuccessTest(BaseWebTest, unittest.TestCase):

    def test_returns_storage_true_if_ok(self):
        response = self.app.get('/__heartbeat__')
        self.assertEqual(response.json['storage'], True)

    def test_returns_cache_true_if_ok(self):
        response = self.app.get('/__heartbeat__')
        self.assertEqual(response.json['cache'], True)

    def test_successful_if_one_heartbeat_is_none(self):
        self.app.app.registry.heartbeats['probe'] = lambda r: None
        response = self.app.get('/__heartbeat__', status=200)
        self.assertEqual(response.json['probe'], None)


class FailureTest(BaseWebTest, unittest.TestCase):

    def test_returns_storage_false_if_ko(self):
        self.app.app.registry.heartbeats['storage'] = lambda r: False
        response = self.app.get('/__heartbeat__', status=503)
        self.assertEqual(response.json['storage'], False)
        self.assertEqual(response.json['cache'], True)

    def test_returns_cache_false_if_ko(self):
        self.app.app.registry.heartbeats['cache'] = lambda r: False
        response = self.app.get('/__heartbeat__', status=503)
        self.assertEqual(response.json['cache'], False)
        self.assertEqual(response.json['storage'], True)

    def test_returns_false_if_heartbeat_times_out(self):
        def sleepy(request):
            import time
            time.sleep(1)
        self.app.app.registry.heartbeats['cache'] = sleepy
        with mock.patch.dict(self.app.app.registry.settings,
                             [('heartbeat_timeout_seconds', 0.1)]):
            response = self.app.get('/__heartbeat__', status=503)
        self.assertEqual(response.json['cache'], False)
        self.assertEqual(response.json['storage'], True)

    def test_returns_false_if_heartbeat_fails(self):
        self.app.app.registry.heartbeats['cache'] = lambda r: 1 / 0
        response = self.app.get('/__heartbeat__', status=503)
        self.assertEqual(response.json['cache'], False)
        self.assertEqual(response.json['storage'], True)


class LoadBalancerHeartbeat(BaseWebTest, unittest.TestCase):
    def test_returns_200_with_empty_body(self):
        resp = self.app.get('/__lbheartbeat__', status=200)
        self.assertEqual(resp.json, {})



import re
import operator
from collections import defaultdict

from kinto.core import utils
from kinto.core.decorators import synchronized
from kinto.core.storage import (
    StorageBase, exceptions,
    DEFAULT_ID_FIELD, DEFAULT_MODIFIED_FIELD, DEFAULT_DELETED_FIELD)
from kinto.core.utils import (COMPARISON, find_nested_value)


def tree():
    return defaultdict(tree)


class MemoryBasedStorage(StorageBase):
    """Abstract storage class, providing basic operations and
    methods for in-memory implementations of sorting and filtering.
    """
    def __init__(self, *args, **kwargs):
        pass

    def initialize_schema(self, dry_run=False):
        # Nothing to do.
        pass

    def strip_deleted_record(self, collection_id, parent_id, record,
                             id_field=DEFAULT_ID_FIELD,
                             modified_field=DEFAULT_MODIFIED_FIELD,
                             deleted_field=DEFAULT_DELETED_FIELD):
        """Strip the record of all its fields expect id and timestamp,
        and set the deletion field value (e.g deleted=True)
        """
        deleted = {}
        deleted[id_field] = record[id_field]
        deleted[modified_field] = record[modified_field]
        deleted[deleted_field] = True
        return deleted

    def set_record_timestamp(self, collection_id, parent_id, record,
                             modified_field=DEFAULT_MODIFIED_FIELD,
                             last_modified=None):
        timestamp = self._bump_timestamp(collection_id, parent_id, record,
                                         modified_field,
                                         last_modified=last_modified)
        record[modified_field] = timestamp
        return record

    def extract_record_set(self, records,
                           filters, sorting, id_field, deleted_field,
                           pagination_rules=None, limit=None):
        """Take the list of records and handle filtering, sorting and
        pagination.

        """
        return extract_record_set(records,
                                  filters=filters,
                                  sorting=sorting,
                                  id_field=id_field,
                                  deleted_field=deleted_field,
                                  pagination_rules=pagination_rules,
                                  limit=limit)


class Storage(MemoryBasedStorage):
    """Storage backend implementation in memory.

    Useful for development or testing purposes, but records are lost after
    each server restart.

    Enable in configuration::

        kinto.storage_backend = kinto.core.storage.memory
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flush()

    def flush(self, auth=None):
        self._store = tree()
        self._cemetery = tree()
        self._timestamps = defaultdict(dict)

    @synchronized
    def collection_timestamp(self, collection_id, parent_id, auth=None):
        ts = self._timestamps[parent_id].get(collection_id)
        if ts is not None:
            return ts
        return self._bump_timestamp(collection_id, parent_id)

    def _bump_timestamp(self, collection_id, parent_id, record=None,
                        modified_field=None, last_modified=None):
        """Timestamp are base on current millisecond.

        .. note ::

            Here it is assumed that if requests from the same user burst in,
            the time will slide into the future. It is not problematic since
            the timestamp notion is opaque, and behaves like a revision number.
        """
        # XXX factorize code from memory and redis backends.
        is_specified = (record is not None and
                        modified_field in record or
                        last_modified is not None)
        if is_specified:
            # If there is a timestamp in the new record, try to use it.
            if last_modified is not None:
                current = last_modified
            else:
                current = record[modified_field]
        else:
            # Otherwise, use a new one.
            current = utils.msec_time()

        # Bump the timestamp only if it's more than the previous one.
        previous = self._timestamps[parent_id].get(collection_id)
        if previous and previous >= current:
            collection_timestamp = previous + 1
        else:
            collection_timestamp = current

        # In case the timestamp was specified, the collection timestamp will
        # be different from the updated timestamp. As such, we want to return
        # the one of the record, and not the collection one.
        if not is_specified or previous == current:
            current = collection_timestamp

        self._timestamps[parent_id][collection_id] = collection_timestamp
        return current

    @synchronized
    def create(self, collection_id, parent_id, record, id_generator=None,
               id_field=DEFAULT_ID_FIELD,
               modified_field=DEFAULT_MODIFIED_FIELD, auth=None, ignore_conflict=False):
        id_generator = id_generator or self.id_generator
        record = {**record}
        if id_field in record:
            # Raise unicity error if record with same id already exists.
            try:
                existing = self.get(collection_id, parent_id, record[id_field])
                if ignore_conflict:
                    return existing
                raise exceptions.UnicityError(id_field, existing)
            except exceptions.RecordNotFoundError:
                pass
        else:
            record[id_field] = id_generator()

        self.set_record_timestamp(collection_id, parent_id, record,
                                  modified_field=modified_field)
        _id = record[id_field]
        self._store[parent_id][collection_id][_id] = record
        self._cemetery[parent_id][collection_id].pop(_id, None)
        return record

    @synchronized
    def get(self, collection_id, parent_id, object_id,
            id_field=DEFAULT_ID_FIELD,
            modified_field=DEFAULT_MODIFIED_FIELD,
            auth=None):
        collection = self._store[parent_id][collection_id]
        if object_id not in collection:
            raise exceptions.RecordNotFoundError(object_id)
        return {**collection[object_id]}

    @synchronized
    def update(self, collection_id, parent_id, object_id, record,
               id_field=DEFAULT_ID_FIELD,
               modified_field=DEFAULT_MODIFIED_FIELD,
               auth=None):
        record = {**record}
        record[id_field] = object_id

        self.set_record_timestamp(collection_id, parent_id, record,
                                  modified_field=modified_field)
        self._store[parent_id][collection_id][object_id] = record
        self._cemetery[parent_id][collection_id].pop(object_id, None)
        return record

    @synchronized
    def delete(self, collection_id, parent_id, object_id,
               id_field=DEFAULT_ID_FIELD, with_deleted=True,
               modified_field=DEFAULT_MODIFIED_FIELD,
               deleted_field=DEFAULT_DELETED_FIELD,
               auth=None, last_modified=None):
        existing = self.get(collection_id, parent_id, object_id)
        # Need to delete the last_modified field of the record.
        del existing[modified_field]

        self.set_record_timestamp(collection_id, parent_id, existing,
                                  modified_field=modified_field,
                                  last_modified=last_modified)
        existing = self.strip_deleted_record(collection_id,
                                             parent_id,
                                             existing)

        # Add to deleted items, remove from store.
        if with_deleted:
            deleted = {**existing}
            self._cemetery[parent_id][collection_id][object_id] = deleted
        self._store[parent_id][collection_id].pop(object_id)
        return existing

    @synchronized
    def purge_deleted(self, collection_id, parent_id, before=None,
                      id_field=DEFAULT_ID_FIELD,
                      modified_field=DEFAULT_MODIFIED_FIELD,
                      auth=None):
        parent_id_match = re.compile(parent_id.replace('*', '.*'))
        by_parent_id = {pid: collections
                        for pid, collections in self._cemetery.items()
                        if parent_id_match.match(pid)}
        num_deleted = 0
        for pid, collections in by_parent_id.items():
            if collection_id is not None:
                collections = {collection_id: collections[collection_id]}
            for collection, colrecords in collections.items():
                if before is None:
                    kept = {}
                else:
                    kept = {key: value for key, value in
                            colrecords.items()
                            if value[modified_field] >= before}
                self._cemetery[pid][collection] = kept
                num_deleted += (len(colrecords) - len(kept))
        return num_deleted

    @synchronized
    def get_all(self, collection_id, parent_id, filters=None, sorting=None,
                pagination_rules=None, limit=None, include_deleted=False,
                id_field=DEFAULT_ID_FIELD,
                modified_field=DEFAULT_MODIFIED_FIELD,
                deleted_field=DEFAULT_DELETED_FIELD,
                auth=None):

        records = _get_objects_by_parent_id(self._store, parent_id, collection_id)

        records, count = self.extract_record_set(records=records,
                                                 filters=filters, sorting=None,
                                                 id_field=id_field, deleted_field=deleted_field)
        deleted = []
        if include_deleted:
            deleted = _get_objects_by_parent_id(self._cemetery, parent_id, collection_id)

        records, count = self.extract_record_set(records=records + deleted,
                                                 filters=filters, sorting=sorting,
                                                 id_field=id_field, deleted_field=deleted_field,
                                                 pagination_rules=pagination_rules, limit=limit)
        return records, count

    @synchronized
    def delete_all(self, collection_id, parent_id, filters=None,
                   sorting=None, pagination_rules=None, limit=None,
                   id_field=DEFAULT_ID_FIELD, with_deleted=True,
                   modified_field=DEFAULT_MODIFIED_FIELD,
                   deleted_field=DEFAULT_DELETED_FIELD,
                   auth=None):
        records = _get_objects_by_parent_id(self._store, parent_id, collection_id, with_meta=True)
        records, count = self.extract_record_set(records=records,
                                                 filters=filters,
                                                 sorting=sorting,
                                                 pagination_rules=pagination_rules, limit=limit,
                                                 id_field=id_field,
                                                 deleted_field=deleted_field)

        deleted = [self.delete(r.pop('__collection_id__'),
                               r.pop('__parent_id__'),
                               r[id_field],
                               id_field=id_field, with_deleted=with_deleted,
                               modified_field=modified_field,
                               deleted_field=deleted_field)
                   for r in records]
        return deleted


def extract_record_set(records, filters, sorting,
                       pagination_rules=None, limit=None,
                       id_field=DEFAULT_ID_FIELD,
                       deleted_field=DEFAULT_DELETED_FIELD):
    """Apply filters, sorting, limit, and pagination rules to the list of
    `records`.

    """
    filtered = list(apply_filters(records, filters or []))
    total_records = len(filtered)

    paginated = {}
    for rule in pagination_rules or []:
        values = list(apply_filters(filtered, rule))
        paginated.update(dict(((x[id_field], x) for x in values)))

    if paginated:
        paginated = paginated.values()
    else:
        paginated = filtered

    sorted_ = apply_sorting(paginated, sorting or [])

    filtered_deleted = len([r for r in sorted_
                            if r.get(deleted_field) is True])

    if limit:
        sorted_ = list(sorted_)[:limit]

    return sorted_, total_records - filtered_deleted


def apply_filters(records, filters):
    """Filter the specified records, using basic iteration.
    """
    operators = {
        COMPARISON.LT: operator.lt,
        COMPARISON.MAX: operator.le,
        COMPARISON.EQ: operator.eq,
        COMPARISON.NOT: operator.ne,
        COMPARISON.MIN: operator.ge,
        COMPARISON.GT: operator.gt,
        COMPARISON.IN: operator.contains,
        COMPARISON.EXCLUDE: lambda x, y: not operator.contains(x, y),
        COMPARISON.LIKE: lambda x, y: re.search(y, x, re.IGNORECASE),
    }
    for record in records:
        matches = True
        for f in filters:
            right = f.value
            if f.field == DEFAULT_ID_FIELD:
                if isinstance(right, int):
                    right = str(right)

            left = find_nested_value(record, f.field)

            if f.operator in (COMPARISON.IN, COMPARISON.EXCLUDE):
                right, left = left, right
            else:
                if left is None:
                    right_is_number = (
                        isinstance(right, (int, float)) and
                        right not in (True, False))
                    if right_is_number:
                        # Python3 cannot compare None to a number.
                        matches = False
                        continue
                    else:
                        left = ''  # To mimic what we do for postgresql.
            matches = matches and operators[f.operator](left, right)
        if matches:
            yield record


def apply_sorting(records, sorting):
    """Sort the specified records, using cumulative python sorting.
    """
    result = list(records)

    if not result:
        return result

    first_record = result[0]

    def column(first, record, name):
        return find_nested_value(record, name, default=float('inf'))

    for sort in reversed(sorting):
        result = sorted(result,
                        key=lambda r: column(first_record, r, sort.field),
                        reverse=(sort.direction < 0))

    return result


def _get_objects_by_parent_id(store, parent_id, collection_id, with_meta=False):
    if parent_id is not None:
        parent_id_match = re.compile("^{}$".format(parent_id.replace('*', '.*')))
        by_parent_id = {pid: collections
                        for pid, collections in store.items()
                        if parent_id_match.match(pid)}
    else:
        by_parent_id = store[parent_id]

    objects = []
    for pid, collections in by_parent_id.items():
        if collection_id is not None:
            collections = {collection_id: collections[collection_id]}
        for collection, colobjects in collections.items():
            for r in colobjects.values():
                if with_meta:
                    objects.append(dict(__collection_id__=collection,
                                        __parent_id__=pid, **r))
                else:
                    objects.append(r)
    return objects


def load_from_config(config):
    return Storage()



import colander
from pyramid.security import NO_PERMISSION_REQUIRED
from cornice.service import get_services

from kinto.core import Service
from kinto.core.api import OpenAPI

swagger = Service(name="swagger", path='/__api__', description="OpenAPI description")


class SwaggerResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


swagger_response_schemas = {
    '200': SwaggerResponseSchema(
        description='Return an OpenAPI description og the running instance.')
}


@swagger.get(permission=NO_PERMISSION_REQUIRED,
             response_schemas=swagger_response_schemas,
             tags=['Utilities'], operation_id='get_openapi_spec')
def swagger_view(request):

    # Only build json once
    try:
        return swagger_view.__json__
    except AttributeError:
        swagger_view.__json__ = OpenAPI(get_services(), request).generate()
        return swagger_view.__json__



from bravado_core.response import validate_response

from .support import SwaggerTest, MINIMALIST_GROUP


class SwaggerGroupResponsesTest(SwaggerTest):

    def test_get_group_200(self):
        response = self.app.get('/buckets/b1/groups/g1',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].get_group
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_group_200(self):
        response = self.app.post_json('/buckets/b1/groups', self.group,
                                      headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].create_group
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_group_201(self):
        response = self.app.post_json('/buckets/b1/groups', MINIMALIST_GROUP,
                                      headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].create_group
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_put_group_200(self):
        response = self.app.put_json('/buckets/b1/groups/g1', MINIMALIST_GROUP,
                                     headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].update_group
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_put_group_201(self):
        response = self.app.put_json('/buckets/b1/groups/g2', MINIMALIST_GROUP,
                                     headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].update_group
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_delete_group_200(self):
        response = self.app.delete('/buckets/b1/groups/g1',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].delete_group
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_get_groups_200(self):
        response = self.app.get('/buckets/b1/groups',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].get_groups
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_delete_groups_200(self):
        response = self.app.delete('/buckets/b1/groups',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Groups'].delete_groups
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)



import mock

from pyramid import testing

from kinto.core.storage import exceptions
from kinto.core.permission import heartbeat
from kinto.core.testing import DummyRequest


class PermissionTest:
    backend = None
    settings = {}

    def setUp(self):
        super().setUp()
        self.permission = self.backend.load_from_config(self._get_config())
        self.permission.initialize_schema()
        self.request = DummyRequest()
        self.client_error_patcher = []

    def _get_config(self):
        """Mock Pyramid config object.
        """
        config = testing.setUp()
        config.add_settings(self.settings)
        return config

    def tearDown(self):
        mock.patch.stopall()
        super().tearDown()
        self.permission.flush()

    def test_backend_error_is_raised_anywhere(self):
        for patch in self.client_error_patcher:
            patch.start()
        calls = [
            (self.permission.flush,),
            (self.permission.add_user_principal, '', ''),
            (self.permission.remove_user_principal, '', ''),
            (self.permission.get_user_principals, ''),
            (self.permission.add_principal_to_ace, '', '', ''),
            (self.permission.remove_principal_from_ace, '', '', ''),
            (self.permission.get_object_permission_principals, '', ''),
            (self.permission.get_object_permissions, ''),
            (self.permission.replace_object_permissions, '', {'write': []}),
            (self.permission.delete_object_permissions, ''),
            (self.permission.get_accessible_objects, []),
            (self.permission.get_authorized_principals, [('*', 'read')]),
        ]
        for call in calls:
            self.assertRaises(exceptions.BackendError, *call)

    def test_initialize_schema_is_idempotent(self):
        self.permission.initialize_schema()
        self.permission.initialize_schema()  # not raising.

    def test_ping_returns_false_if_unavailable(self):
        ping = heartbeat(self.permission)
        for patch in self.client_error_patcher:
            patch.start()
        self.assertFalse(ping(self.request))

    def test_ping_returns_true_if_available(self):
        ping = heartbeat(self.permission)
        self.assertTrue(ping(self.request))

    def test_ping_returns_false_if_unavailable_in_readonly_mode(self):
        self.request.registry.settings['readonly'] = 'true'
        ping = heartbeat(self.permission)
        with mock.patch.object(self.permission, 'get_user_principals',
                               side_effect=exceptions.BackendError("Boom!")):
            self.assertFalse(ping(self.request))

    def test_ping_returns_true_if_available_in_readonly_mode(self):
        self.request.registry.settings['readonly'] = 'true'
        ping = heartbeat(self.permission)
        self.assertTrue(ping(self.request))

    def test_ping_logs_error_if_unavailable(self):
        for patch in self.client_error_patcher:
            patch.start()
        ping = heartbeat(self.permission)

        with mock.patch('kinto.core.permission.logger.exception') as \
                exc_handler:
            self.assertFalse(ping(self.request))

        self.assertTrue(exc_handler.called)

    def test_can_add_a_principal_to_a_user(self):
        user_id = 'foo'
        principal = 'bar'
        self.permission.add_user_principal(user_id, principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEquals(retrieved, {principal})

    def test_add_twice_a_principal_to_a_user_add_it_once(self):
        user_id = 'foo'
        principal = 'bar'
        self.permission.add_user_principal(user_id, principal)
        self.permission.add_user_principal(user_id, principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEquals(retrieved, {principal})

    def test_can_remove_a_principal_for_an_unknown_user(self):
        self.permission.remove_user_principal('foo', 'bar')

    def test_can_remove_a_principal_for_a_user(self):
        user_id = 'foo'
        principal = 'bar'
        principal2 = 'foobar'
        self.permission.add_user_principal(user_id, principal)
        self.permission.add_user_principal(user_id, principal2)
        self.permission.remove_user_principal(user_id, principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEquals(retrieved, {principal2})

    def test_can_remove_a_unexisting_principal_to_a_user(self):
        user_id = 'foo'
        principal = 'bar'
        principal2 = 'foobar'
        self.permission.add_user_principal(user_id, principal2)
        self.permission.remove_user_principal(user_id, principal)
        self.permission.remove_user_principal(user_id, principal2)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEquals(retrieved, set())

    def test_can_remove_principal_from_every_users(self):
        user_id1 = 'foo1'
        user_id2 = 'foo2'
        principal1 = 'bar'
        principal2 = 'foobar'
        self.permission.add_user_principal(user_id1, principal1)
        self.permission.add_user_principal(user_id2, principal1)
        self.permission.add_user_principal(user_id2, principal2)
        self.permission.remove_principal(principal1)
        self.permission.remove_principal('unknown')

        retrieved = self.permission.get_user_principals(user_id1)
        self.assertEquals(retrieved, set())
        retrieved = self.permission.get_user_principals(user_id2)
        self.assertEquals(retrieved, {principal2})

    def test_authenticated_is_returned_for_everybody(self):
        user_id = 'foo'
        principal = 'bar'
        self.permission.add_user_principal('system.Authenticated', principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEquals(retrieved, {principal})

    #
    # get_object_permission_principals()
    #

    def test_can_add_a_principal_to_an_object_permission(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        self.permission.add_principal_to_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(
            object_id, permission)
        self.assertEquals(retrieved, {principal})

    def test_add_twice_a_principal_to_an_object_permission_add_it_once(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        self.permission.add_principal_to_ace(object_id, permission, principal)
        self.permission.add_principal_to_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(
            object_id, permission)
        self.assertEquals(retrieved, {principal})

    def test_can_remove_a_principal_from_an_object_permission(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        principal2 = 'foobar'
        self.permission.add_principal_to_ace(object_id, permission, principal)
        self.permission.add_principal_to_ace(object_id, permission, principal2)
        self.permission.remove_principal_from_ace(object_id, permission,
                                                  principal)
        retrieved = self.permission.get_object_permission_principals(
            object_id, permission)
        self.assertEquals(retrieved, {principal2})

    def test_principals_is_empty_if_no_permission(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        self.permission.add_principal_to_ace(object_id, permission, principal)
        self.permission.remove_principal_from_ace(object_id, permission,
                                                  principal)
        retrieved = self.permission.get_object_permission_principals(
            object_id, permission)
        self.assertEquals(retrieved, set())

    def test_can_remove_an_unexisting_principal_to_an_object_permission(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        principal2 = 'foobar'
        self.permission.add_principal_to_ace(object_id, permission, principal2)
        self.permission.remove_principal_from_ace(object_id, permission,
                                                  principal)
        retrieved = self.permission.get_object_permission_principals(
            object_id, permission)
        self.assertEquals(retrieved, {principal2})

    #
    # check_permission()
    #

    def test_check_permission_returns_true_for_userid(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        self.permission.add_principal_to_ace(object_id, permission, principal)
        check_permission = self.permission.check_permission(
            {principal},
            [(object_id, permission)])
        self.assertTrue(check_permission)

    def test_check_permission_returns_true_for_userid_group(self):
        object_id = 'foo'
        permission = 'write'
        group_id = 'bar'
        user_id = 'foobar'
        self.permission.add_user_principal(user_id, group_id)
        self.permission.add_principal_to_ace(object_id, permission, group_id)
        check_permission = self.permission.check_permission(
            {user_id, group_id},
            [(object_id, permission)])
        self.assertTrue(check_permission)

    def test_check_permission_returns_true_for_object_inherited(self):
        object_id = 'foo'
        user_id = 'bar'
        self.permission.add_principal_to_ace(object_id, 'write', user_id)
        check_permission = self.permission.check_permission(
            {user_id},
            [(object_id, 'write'), (object_id, 'read')])
        self.assertTrue(check_permission)

    def test_check_permissions_handles_empty_set(self):
        principal = 'bar'
        permits = self.permission.check_permission({principal}, [])
        self.assertFalse(permits)

    def test_check_permission_return_false_for_unknown_principal(self):
        object_id = 'foo'
        permission = 'write'
        principal = 'bar'
        check_permission = self.permission.check_permission(
            {principal},
            [(object_id, permission)])
        self.assertFalse(check_permission)

    #
    # get_authorized_principals()
    #

    def test_get_authorized_principals_inherit_principals(self):
        object_id = 'foo'
        user_id = 'bar'
        self.permission.add_principal_to_ace(object_id, 'write', user_id)
        principals = self.permission.get_authorized_principals(
            [(object_id, 'write'), (object_id, 'read')])
        self.assertEquals(principals, {user_id})

    def test_get_authorized_principals_handles_empty_set(self):
        principals = self.permission.get_authorized_principals([])
        self.assertEquals(principals, set())

    #
    # get_accessible_objects()
    #

    def test_accessible_objects(self):
        self.permission.add_principal_to_ace('id1', 'write', 'user1')
        self.permission.add_principal_to_ace('id1', 'record:create', 'group')
        self.permission.add_principal_to_ace('id2', 'read', 'user1')
        self.permission.add_principal_to_ace('id2', 'read', 'user2')
        self.permission.add_principal_to_ace('id3', 'write', 'user2')
        per_object_ids = self.permission.get_accessible_objects(
            ['user1', 'group'])
        self.assertEquals(sorted(per_object_ids.keys()), ['id1', 'id2'])
        self.assertEquals(per_object_ids['id1'],
                          set(['write', 'record:create']))
        self.assertEquals(per_object_ids['id2'], set(['read']))

    def test_accessible_objects_supports_empty_list(self):
        per_object_ids = self.permission.get_accessible_objects(['user1', 'group'], [])
        self.assertEquals(per_object_ids, {})

    def test_accessible_objects_from_permission(self):
        self.permission.add_principal_to_ace('id1', 'write', 'user1')
        self.permission.add_principal_to_ace('id1', 'read', 'user1')
        self.permission.add_principal_to_ace('id1', 'read', 'group')
        self.permission.add_principal_to_ace('id2', 'write', 'user1')
        self.permission.add_principal_to_ace('id2', 'read', 'user2')
        self.permission.add_principal_to_ace('id2', 'read', 'group')
        self.permission.add_principal_to_ace('id3', 'read', 'user2')
        per_object_ids = self.permission.get_accessible_objects(
            ['user1', 'group'],
            [('*', 'read')])
        self.assertEquals(sorted(per_object_ids.keys()), ['id1', 'id2'])

    def test_accessible_objects_with_pattern(self):
        self.permission.add_principal_to_ace('/url1/id', 'write', 'user1')
        self.permission.add_principal_to_ace('/url2/id', 'write', 'user1')
        per_object_ids = self.permission.get_accessible_objects(
            ['user1'],
            [('*url1*', 'write')])
        self.assertEquals(sorted(per_object_ids.keys()), ['/url1/id'])

    def test_accessible_objects_with_pattern_matches_whole_id(self):
        self.permission.add_principal_to_ace('/url1/id', 'write', 'user1')
        self.permission.add_principal_to_ace('/url1/id/sub', 'write', 'user1')
        self.permission.add_principal_to_ace('/a/url1/id', 'write', 'user1')
        per_object_ids = self.permission.get_accessible_objects(
            ['user1'],
            [('/url1/*', 'write')],
            with_children=False)
        self.assertEquals(sorted(per_object_ids.keys()), ['/url1/id'])

    def test_accessible_objects_several_bound_permissions(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/2', 'read', 'user1')
        self.permission.add_principal_to_ace('/url/_/id/2', 'read', 'user1')
        per_object_ids = self.permission.get_accessible_objects(
            ['user1'],
            [('/url/a/id/*', 'read'),
             ('/url/a/id/*', 'write')])
        self.assertEquals(sorted(per_object_ids.keys()),
                          ['/url/a/id/1', '/url/a/id/2'])

    def test_accessible_objects_without_match(self):
        self.permission.add_principal_to_ace('/url/a', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/b/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/2', 'read', 'user1')
        self.permission.add_principal_to_ace('/url/b/id/2', 'read', 'user1')
        per_object_ids = self.permission.get_accessible_objects(
            ['user1'],
            [('/url/a', 'write'),
             ('/url/a', 'read'),
             ('/url/a/id/*', 'write'),
             ('/url/a/id/*', 'read')])
        self.assertEquals(sorted(per_object_ids.keys()),
                          ['/url/a', '/url/a/id/1', '/url/a/id/2'])

    #
    # get_object_permissions()
    #

    def test_object_permissions_return_all_object_acls(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user2')
        self.permission.add_principal_to_ace('/url/a/id/1', 'read', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/1', 'obj:del', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1/sub', 'create', 'me')
        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertDictEqual(permissions, {
            "write": {"user1", "user2"},
            "read": {"user3"},
            "obj:del": {"user1"}
        })

    def test_object_permissions_return_listed_object_acls(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user2')
        self.permission.add_principal_to_ace('/url/a/id/1', 'read', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/1', 'create', 'user1')
        object_permissions = self.permission.get_object_permissions(
            '/url/a/id/1', ['write', 'read'])
        self.assertDictEqual(object_permissions, {
            "write": {"user1", "user2"},
            "read": {"user3"}
        })

    def test_objects_permissions_returns_empty_if_unknown(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/3', 'read', 'user3')
        objects_permissions = self.permission.get_objects_permissions([
            '/url/a/id/1', '/abc', '/url/a/id/3'])
        self.assertEqual(objects_permissions, [
            {"write": {"user1"}}, {}, {"read": {"user3"}}])

    def test_object_permissions_return_empty_dict(self):
        self.assertDictEqual(self.permission.get_object_permissions('abc'), {})

    def test_replace_object_permission_replace_all_given_sets(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user2')
        self.permission.add_principal_to_ace('/url/a/id/1', 'read', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/1', 'update', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'obj:del', 'user1')

        self.permission.replace_object_permissions('/url/a/id/1', {
            "write": ["user1"],
            "read": ["user2"],
            "update": [],
            "obj:del": ["user1"],
            "new": ["user3"]
        })

        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertDictEqual(permissions, {
            "write": {"user1"},
            "read": {"user2"},
            "obj:del": {"user1"},
            "new": {"user3"}
        })

    def test_replace_object_permission_only_replace_given_sets(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user2')
        self.permission.add_principal_to_ace('/url/a/id/1', 'read', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/1', 'obj:del', 'user1')

        self.permission.replace_object_permissions('/url/a/id/1', {
            "write": ["user1"],
            "new": set(["user2"])
        })

        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertDictEqual(permissions, {
            "write": {"user1"},
            "read": {"user3"},
            "new": {"user2"},
            "obj:del": {"user1"}
        })

    def test_replace_object_permission_supports_empty_input(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.replace_object_permissions('/url/a/id/1', {})
        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertDictEqual(permissions, {
            "write": {"user1"}
        })

    def test_replace_object_permission_supports_duplicated_entries(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.replace_object_permissions('/url/a/id/1', {
            "write": ["user1", "user1"]
        })
        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertDictEqual(permissions, {
            "write": {"user1"}
        })

    def test_replace_object_permission_supports_empty_list(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.replace_object_permissions('/url/a/id/1', {
            "write": set()
        })
        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertEqual(len(permissions), 0)

    def test_replace_object_permission_supports_empty_list_to_new_object(self):
        self.permission.replace_object_permissions('/url/a/id/1', {
            "write": set()
        })
        permissions = self.permission.get_object_permissions('/url/a/id/1')
        self.assertEqual(len(permissions), 0)

    def test_delete_object_permissions_remove_all_given_objects_acls(self):
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user2')
        self.permission.add_principal_to_ace('/url/a/id/1', 'read', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/1', 'create', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/2', 'create', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/3', 'create', 'user4')

        self.permission.delete_object_permissions('/url/a/id/1',
                                                  '/url/a/id/2')

        self.assertDictEqual(self.permission.get_object_permissions(
            '/url/a/id/1'), {})
        self.assertDictEqual(self.permission.get_object_permissions(
            '/url/a/id/2'), {})
        self.assertDictEqual(self.permission.get_object_permissions(
            '/url/a/id/3'), {"create": {"user4"}})

    def test_delete_object_permissions_supports_empty_list(self):
        self.permission.delete_object_permissions()  # Not failing

    def test_delete_object_permissions_supports_pattern_matching(self):
        self.permission.add_principal_to_ace('/url/b/id/1', 'write', 'user1')
        self.permission.add_principal_to_ace('/url/a/id/1', 'write', 'user2')
        self.permission.add_principal_to_ace('/url/a/id/1', 'read', 'user3')
        self.permission.add_principal_to_ace('/url/a/id/3', 'create', 'user4')

        self.permission.delete_object_permissions('/url/a*')

        self.assertDictEqual(self.permission.get_object_permissions('/url/a/id/1'), {})
        self.assertDictEqual(
            self.permission.get_object_permissions('/url/b/id/1'),
            {'write': {'user1'}})
